@@ -4,6 +4,7 @@ This software is distributed under the GNU General Public License.
 See the file COPYING for details.
 */
 
+#include <assert.h>
 #include <stdio.h>
 #include <string.h>
 #include <strings.h>
@@ -422,6 +423,12 @@ struct rmonitor_mem_info *rmonitor_get_map_info(FILE *fmem, int rewind_flag) {
 	return NULL;
 }
 
+static double rmonitor_mem_info_priority(void *item) {
+	assert(item);
+	struct rmonitor_mem_info *i = item;
+	return -1*(i->map_start);
+}
+
 int rmonitor_get_mmaps_usage(pid_t pid, struct hash_table *maps)
 {
 	// /dev/proc/[pid]/smaps:
@@ -474,7 +481,7 @@ int rmonitor_get_mmaps_usage(pid_t pid, struct hash_table *maps)
 			hash_table_insert(maps, info->map_name, infos);
 		}
 
-		list_push_priority(infos, info, -1*(info->map_start));
+		list_push_priority(infos, rmonitor_mem_info_priority, info);
 	}
 
 	fclose(fmem);
