@@ -39,6 +39,9 @@ See the file COPYING for details.
 #include "url_encode.h"
 #include "md5.h"
 #include "disk_alloc.h"
+#include "hash_table.h"
+#include "pattern.h"
+#include "gpu_info.h"
 
 #include <unistd.h>
 #include <dirent.h>
@@ -199,6 +202,9 @@ static struct list   *procs_waiting = NULL;
 // These are additional pointers into procs_table.
 static struct itable *procs_complete = NULL;
 
+//User specified features this worker provides.
+static struct hash_table *features = NULL;
+
 static int results_to_be_sent_msg = 0;
 
 static timestamp_t total_task_execution_time = 0;
@@ -294,7 +300,7 @@ void measure_worker_resources()
 	work_queue_resources_measure_locally(r,workspace);
 
 	if(worker_mode == WORKER_MODE_FOREMAN) {
-		aggregate_workers_resources(foreman_q, total_resources);
+		aggregate_workers_resources(foreman_q, total_resources, features);
 	} else {
 		if(manual_cores_option > 0)
 			r->cores.total = manual_cores_option;
@@ -327,6 +333,21 @@ void measure_worker_resources()
 	last_resources_measurement = time(0);
 }
 
+/*
+Send a message to the master with user defined features.
+*/
+static void send_features(struct link *master) {
+	char *f;
+	void *dummy;
+	hash_table_firstkey(features);
+
+	char fenc[WORK_QUEUE_LINE_MAX];
+	while(hash_table_nextkey(features, &f, &dummy)) {
+		url_encode(f, fenc, WORK_QUEUE_LINE_MAX);
+		send_master_message(master, "feature %s\n", fenc);
+	}
+}
+
 
 /*
 Send a message to the master with my current resources.
@@ -420,6 +441,7 @@ static void report_worker_ready( struct link *master )
 	domain_name_cache_guess(hostname);
 	send_master_message(master,"workqueue %d %s %s %s %d.%d.%d\n",WORK_QUEUE_PROTOCOL_VERSION,hostname,os_name,arch_name,CCTOOLS_VERSION_MAJOR,CCTOOLS_VERSION_MINOR,CCTOOLS_VERSION_MICRO);
 	send_master_message(master, "info worker-id %s\n", worker_id);
+	send_features(master);
 	send_keepalive(master, 1);
 }
 
@@ -1397,18 +1419,28 @@ static int handle_master(struct link *master) {
 	char path[WORK_QUEUE_LINE_MAX];
 	int64_t length;
 	int64_t taskid = 0;
-	int flags = WORK_QUEUE_NOCACHE;
 	int mode, r, n;
 
 	if(recv_master_message(master, line, sizeof(line), idle_stoptime )) {
 		if(sscanf(line,"task %" SCNd64, &taskid)==1) {
 			r = do_task(master, taskid,time(0)+active_timeout);
-		} else if((n = sscanf(line, "put %s %" SCNd64 " %o %d", filename, &length, &mode, &flags)) >= 3) {
-			if(path_within_dir(filename, workspace)) {
-				r = do_put(master, filename, length, mode);
-				reset_idle_timer();
+		} else if(string_prefix_is(line, "put ")) {
+			char *f = NULL, *l = NULL, *m = NULL, *g = NULL;
+			if(pattern_match(line, "^put (.+) (%d+) ([0-7]+) (%d+)$", &f, &l, &m, &g) >= 0) {
+				strncpy(filename, f, WORK_QUEUE_LINE_MAX); free(f);
+				length = strtoll(l, 0, 10); free(l);
+				mode   = atoi(m);           free(m);
+				free(g); //flags are not used anymore.
+				
+				if(path_within_dir(filename, workspace)) {
+					r = do_put(master, filename, length, mode);
+					reset_idle_timer();
+				} else {
+					debug(D_WQ, "Path - %s is not within workspace %s.", filename, workspace);
+					r = 0;
+				}
 			} else {
-				debug(D_WQ, "Path - %s is not within workspace %s.", filename, workspace);
+				debug(D_WQ, "Malformed put message.");
 				r = 0;
 			}
 		} else if(sscanf(line, "url %s %" SCNd64 " %o", filename, &length, &mode) == 3) {
@@ -2185,6 +2217,7 @@ static void show_help(const char *cmd)
 	printf( " %-30s Manually set the amount of memory (in MB) reported by this worker.\n", "--memory=<mb>           ");
 	printf( " %-30s Manually set the amount of disk (in MB) reported by this worker.\n", "--disk=<mb>");
 	printf( " %-30s Use loop devices for task sandboxes (default=disabled, requires root access).\n", "--disk-allocation");
+	printf( " %-30s Specifies a user-defined feature the worker provides. May be specified several times.\n", "--feature");
 	printf( " %-30s Set the maximum number of seconds the worker may be active. (in s).\n", "--wall-time=<s>");
 	printf( " %-30s Forbid the use of symlinks for cache management.\n", "--disable-symlinks");
 	printf(" %-30s Single-shot mode -- quit immediately after disconnection.\n", "--single-shot");
@@ -2199,7 +2232,7 @@ enum {LONG_OPT_DEBUG_FILESIZE = 256, LONG_OPT_VOLATILITY, LONG_OPT_BANDWIDTH,
 	  LONG_OPT_DISK, LONG_OPT_GPUS, LONG_OPT_FOREMAN, LONG_OPT_FOREMAN_PORT, LONG_OPT_DISABLE_SYMLINKS,
 	  LONG_OPT_IDLE_TIMEOUT, LONG_OPT_CONNECT_TIMEOUT, LONG_OPT_RUN_DOCKER, LONG_OPT_RUN_DOCKER_PRESERVE,
 	  LONG_OPT_BUILD_FROM_TAR, LONG_OPT_SINGLE_SHOT, LONG_OPT_WALL_TIME, LONG_OPT_DISK_ALLOCATION,
-	  LONG_OPT_MEMORY_THRESHOLD};
+	  LONG_OPT_MEMORY_THRESHOLD, LONG_OPT_FEATURE};
 
 static const struct option long_options[] = {
 	{"advertise",           no_argument,        0,  'a'},
@@ -2243,6 +2276,7 @@ static const struct option long_options[] = {
 	{"docker",              required_argument,  0,  LONG_OPT_RUN_DOCKER},
 	{"docker-preserve",     required_argument,  0,  LONG_OPT_RUN_DOCKER_PRESERVE},
 	{"docker-tar",          required_argument,  0,  LONG_OPT_BUILD_FROM_TAR},
+	{"feature",            required_argument,  0,  LONG_OPT_FEATURE},
 	{0,0,0,0}
 };
 
@@ -2259,6 +2293,8 @@ int main(int argc, char *argv[])
 	char *foreman_stats_filename = NULL;
 	char * catalog_hosts = CATALOG_HOST;
 
+	features = hash_table_create(4, 0);
+
 	worker_start_time = time(0);
 
 	set_worker_id();
@@ -2474,6 +2510,9 @@ int main(int argc, char *argv[])
 			disk_allocation = 1;
 			break;
 		}
+		case LONG_OPT_FEATURE:
+			hash_table_insert(features, optarg, (void **) 1);
+			break;
 		default:
 			show_help(argv[0]);
 			return 1;
@@ -2521,6 +2560,12 @@ int main(int argc, char *argv[])
 		}
 	}
 
+	//Check GPU name
+	char *gpu_name = gpu_name_get();
+	if(gpu_name) {
+		hash_table_insert(features, gpu_name, (void **) 1);
+	}
+
 	signal(SIGTERM, handle_abort);
 	signal(SIGQUIT, handle_abort);
 	signal(SIGINT, handle_abort);
